// Bug.m - slaves

#import "Bug.h"
#import <random.h>


@implementation Bug

float inp[4];

- initWeights  // initializes ANN with random weights in [-1, 1]
{
  int from,to;

  for (from = 0; from < 4; from++)
   for (to = 0; to < 4; to++)
    weights[from][to] = (float)[uniformDblRand getDoubleWithMin: -1.0L withMax: 1.0L];

  return self;
}

- setLifeTime: (int)time
{
  lifetime = time;
  age = 0;

  return self;
}

- setLearningConstants: (float)delta_big: (float)delta_small;
{
  DELTA = delta_big;    // DELTA is used when the bug finds food
  delta = delta_small;  // otherwise delta is used as a reinforcement signal

  return self;
}

- setWorld: w Objs: o
{
   world = w;
   objs = o;

   return self;
}

- createEnd
{
   [super createEnd];

   worldXSize = [world getSizeX];
   worldYSize = [world getSizeY];

   return self;
}

- setX: (int)x Y: (int)y
{
  xPos = x;
  yPos = y;

  return self;
}

- getX: (int*)x Y: (int*)y
{
  *x = xPos;
  *y = yPos;
  return self; 
}

// returns 0 if position is free otherwise 1
- (int) okAtX: (int)x Y: (int)y
{
  if ((x >= 0) && (y >= 0) && (x < worldXSize) && (y < worldYSize)
  && ([objs getValueAtX: x Y: y] != 1)
  && ([world getObjectAtX: x Y: y] == nil))
   return 0;

   else return 1;
}

// returns the number of obstacles the bug can see around it
- (int) see_obst
{
  return [self okAtX:xPos Y:yPos-1] + [self okAtX:xPos-1 Y:yPos]
       + [self okAtX:xPos Y:yPos+1] + [self okAtX:xPos+1 Y:yPos];
}

// seed food at a random position
- seedRandomFood
{
  int x,y;

  x = [uniformIntRand getIntegerWithMin: 0 withMax: worldXSize-1];
  y = [uniformIntRand getIntegerWithMin: 0 withMax: worldYSize-1];
  while (([objs getValueAtX: x Y: y] != 0) || ([world getObjectAtX: x Y: y] != nil)) {
    x++; if (x == worldXSize) {
      x = 0; y++; if (y == worldYSize) y = 0;
    }
  }
  [objs putValue: 2 atX: x Y: y];  // seed new food

  return self;
}

- moveToX: (int)x Y: (int)y
{
  int obst_avoided, i, j;

//  obst_avoided = [self see_obst];

  if ([self okAtX: x Y: y] == 0) {  // everything's ok ?

/*    // giving reinforcement signals for avoiding obstacles is bad for learning...
    for (i = 0; i < 4; i++) {
      weights[i][winner] += inp[i]*(float)(obst_avoided * delta);
    } 
*/
    [world putObject: nil atX: xPos Y: yPos];  // move bug to new position
    xPos = x;
    yPos = y;
    [world putObject: self atX: xPos Y: yPos];

    // bug found food  
    if ([objs getValueAtX: x Y: y] == 2) {
      [objs putValue: 0 atX: x Y: y];  // eat it
      lifetime += 100;  // increase life time
      [self seedRandomFood];  //seed food for every eaten item
      haveEaten++;

      for (i = 0; i < 4; i++) 
        weights[i][winner] += inp[i]*DELTA;
    } 
  }

  // bug crashed
  else for (i = 0; i < 4; i++) {
    weights[i][winner] -= inp[i]*delta;
    for (j = 0; j < 4; j++) if (j != winner) 
      weights[i][j] += inp[i]*delta;
  }
}

- step_left
{
  [self moveToX: xPos-1 Y: yPos];
  return self;
}
- step_right
{
  [self moveToX: xPos+1 Y: yPos];
  return self;
}
- step_up
{
  [self moveToX: xPos Y: yPos-1];
  return self;
}
- step_down
{
  [self moveToX: xPos Y: yPos+1];
  return self;
}

- step: (float[4])input_array
{
  int from, to, i;
  float outp[4];

  if (lifetime > 0) {
  for (i = 0; i < 4; i++) inp[i] = input_array[i]; 

  printf("input:   %f %f %f %f\n",inp[0],inp[1],inp[2],inp[3]);

/*  for (i = 0; i < 4; i++) {
    printf("w%i: %f %f %f %f\n",i,weights[0][i],weights[1][i],weights[2][i],weights[3][i]);
  }
*/

  // compute output vector from input vector
  for (to = 0; to < 4; to++) {
    outp[to] = 0.0L; 
    for (from = 0; from < 4; from++)
     outp[to] += inp[from] * weights[from][to];
  }

  winner = 0;
  for (i = 1; i < 4; i++)
   if (outp[i] > outp[winner]) winner = i;

  if (winner == 0) [self step_up]; else
  if (winner == 1) [self step_right]; else
  if (winner == 2) [self step_down]; else
                   [self step_left];

  printf("output:  %f %f %f %f\n",outp[0],outp[1],outp[2],outp[3]);

  lifetime--;  // loose one point of life time
  age++;  // increase age
  
  }
  else {       // remove object from display
    printf("(died after %i steps)\n",age);
    [world putObject: nil atX: xPos Y: yPos];

    // R.I.P.
    if (xPos > 0) [objs putValue: 3 atX: xPos-1 Y: yPos];
    if (xPos < worldXSize-1) [objs putValue: 3 atX: xPos+1 Y: yPos];
    if (yPos > 0) [objs putValue: 3 atX: xPos Y: yPos-1];
    if (yPos < worldYSize-1) [objs putValue: 3 atX: xPos Y: yPos+1];
    if (yPos < worldYSize-2) [objs putValue: 3 atX: xPos Y: yPos+2];
 }
  return self;
}

- drawSelfOn: (id <Raster>)r
{
  [r drawPointX: xPos Y: yPos Color: 3];
  return self;
}
  
@end

