// Master.m

#import "Master.h"
#import <activity.h>
#import <collections.h>
#import <random.h>


@implementation Master

- setNavPoint: bug_nr  // selects a random food destination
{
  // find food for bug

  navToX = [uniformIntRand getIntegerWithMin: 0 withMax: worldXSize-1];
  navToY = [uniformIntRand getIntegerWithMin: 0 withMax: worldYSize-1];
  while ([objs getValueAtX: navToX Y: navToY] != 2) {
    navToX++; if (navToX == worldXSize) {
      navToX = 0; navToY++; if (navToY == worldYSize) navToY = 0;
    }
  }

  [[navList atOffset: bug_nr] setDestX: navToX Y: navToY];

  printf("Bug %i - navigating to %i,%i\n",bug_nr,navToX,navToY);

  return self;
}

- checkNavPoint: bug_nr  // is still food at our destination ?
{
  [[navList atOffset: bug_nr] getDestX: (&navToX) Y: (&navToY)];

  if ([objs getValueAtX: navToX Y: navToY] != 2)  // there is no more food (eaten by another bug)
   [self setNavPoint: bug_nr];  // -> find new one

  return self;
}

- setWorld: w Objs: o
{
   world = w;
   objs = o;

   return self;
}

- setBugList: b
{
  bugList = b;

  return self;
}

- setNavList: n
{
  int i;

  navList = n;

  // set initial food destinations
  for (i = 0; i < [navList getCount]; i++) {
    [self setNavPoint: i];
    [[navList atOffset: i] setOldX: -1 Y: -1];  // outside the world
  }

  return self;
}

- createEnd
{
   [super createEnd];

   worldXSize = [world getSizeX];
   worldYSize = [world getSizeY];

   lifecycle = 0;

   return self;
}

- setX: (int)x Y: (int)y
{
  xPos = x;
  yPos = y;

  return self;
}

- (int) okAtX: (int)x Y: (int)y  // can we go there ?
{
  int value = 0;  // 0 means ok

  if ( (x < 0) || (y < 0) || (x >= worldXSize) || (y >= worldYSize) 
   || ([objs getValueAtX: x Y: y] == 1)
   || ([world getObjectAtX: x Y: y] != nil)
   || ( (x == xPos_old) && (y == yPos_old) ) )
   value = 1;  // don't go there

  return value;
}

- tell_to_go_up
{
  input[0] = 1;
  return self;
}

- tell_to_go_right
{
  input[1] = 1;
  return self;
}

- tell_to_go_down
{
  input[2] = 1;
  return self;
}

- tell_to_go_left
{
  input[3] = 1;
  return self;
}

- speak
{
  int x,y, i;
  id  bug,nav;

  printf("\nLife cycle: %i\n",lifecycle);
  lifecycle++;

  for (i = 0; i < [bugList getCount]; i++) {

  // fill input array with 0s (no direction currently set)
  for (x = 0; x < 4; x++) input[x] = 0.0;

  // get bug destination
  nav = [navList atOffset: i];
  [self checkNavPoint: i];  // checks destination and saves it in navToX and navToY

  // get bug position
  bug = [bugList atOffset: i];
  [bug getX: (&x) Y: (&y)];
  printf("Bug %i: %i,%i (Dest: %i,%i)\n",i,x,y,navToX,navToY);

  // get bug's old position to avoid walking the same way forwards and backwards
  [nav getOldX: (&xPos_old) Y: (&yPos_old)];

  // lead bug to destination
  if (y > navToY) {  // send it up
    if ([self okAtX: x Y: y-1] == 0) [self tell_to_go_up];
    else if ((x > navToX) && ([self okAtX: x-1 Y: y] == 0)) [self tell_to_go_left];
    else if ([self okAtX: x+1 Y: y] == 0) [self tell_to_go_right];
    else if ([self okAtX: x-1 Y: y] == 0) [self tell_to_go_left];
    else if ([self okAtX: x Y: y+1] == 0) [self tell_to_go_down];
  }
  else if (x > navToX) {   // send it left
    if ([self okAtX: x-1 Y: y] == 0) [self tell_to_go_left];
    else if ((y < navToY) && ([self okAtX: x Y: y+1] == 0)) [self tell_to_go_down];
    else if ([self okAtX: x Y: y-1] == 0) [self tell_to_go_up];
    else if ([self okAtX: x Y: y+1] == 0) [self tell_to_go_down];
    else if ([self okAtX: x+1 Y: y] == 0) [self tell_to_go_right];
  }
  else if (y < navToY) {   // send it down
    if ([self okAtX: x Y: y+1] == 0) [self tell_to_go_down];
    else if ((x < navToX) && ([self okAtX: x+1 Y: y] == 0)) [self tell_to_go_right];
    else if ([self okAtX: x-1 Y: y] == 0) [self tell_to_go_left];
    else if ([self okAtX: x+1 Y: y] == 0) [self tell_to_go_right];
    else if ([self okAtX: x Y: y-1] == 0) [self tell_to_go_up];
  }
  else if (x < navToX) {   // send it right
    if ([self okAtX: x+1 Y: y] == 0) [self tell_to_go_right];
    else if ((y > navToY) && ([self okAtX: x Y: y-1] == 0)) [self tell_to_go_up];
    else if ([self okAtX: x Y: y+1] == 0) [self tell_to_go_down];
    else if ([self okAtX: x Y: y-1] == 0) [self tell_to_go_up];
    else if ([self okAtX: x-1 Y: y] == 0) [self tell_to_go_left];
  }

  [bug step:input];

  // remember old position
  [nav setOldX: x Y: y];

  }

  return self;
}

- drawSelfOn: (id <Raster>)r
{
  [r drawPointX: xPos Y: yPos Color: 4];
  return self;
}
  
@end

